package com.famoco.smartcardapi_sample_app;

import android.os.Bundle;
import android.util.Log;
import android.view.View;
import android.widget.AdapterView;
import android.widget.ArrayAdapter;
import android.widget.Button;
import android.widget.EditText;
import android.widget.Spinner;
import android.widget.TextView;
import android.widget.Toast;

import androidx.appcompat.app.AppCompatActivity;

import com.famoco.smartcardapi_sample_app.utils.ByteArrayUtils;

import org.simalliance.openmobileapi.Channel;
import org.simalliance.openmobileapi.Reader;
import org.simalliance.openmobileapi.SEService;
import org.simalliance.openmobileapi.Session;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Locale;
import java.util.NoSuchElementException;

public class MainActivity extends AppCompatActivity implements SEService.CallBack {

    private static final String TAG = MainActivity.class.getSimpleName();

    // APDU constants
    private static final String SELECTION_NONE = "Custom";

    private static final String COMMAND_SAM_GET_VERSION = "SAM_AV2_GetVersion";
    private static final String APDU_SAM_GET_VERSION = "80 60 00 00 00";

    private static final String COMMAND_SELECT_CARD_MANAGER = "SELECT eSE ISD";
    private static final String APDU_SELECT_CARD_MANAGER = "00 A4 04 00 08 A0 00 00 01 51 00 00 00";

    private static final String COMMAND_GET_CPLC = "GET eSE CPLC";
    private static final String APDU_GET_CPLC = "80 CA 9F 7F 00";

    private static final String COMMAND_GET_CHALLENGE = "SAM_AV2_GetChallenge";
    private static final String APDU_GET_CHALLENGE = "80 84 00 00 0F";

    private static final String[] COMMAND_ARRAY = new String[] {
            COMMAND_SELECT_CARD_MANAGER,
            COMMAND_SAM_GET_VERSION,
            COMMAND_GET_CPLC,
            COMMAND_GET_CHALLENGE,
            SELECTION_NONE,
    };

    private static final String[] APDU_ARRAY = new String[] {
            APDU_SELECT_CARD_MANAGER,
            APDU_SAM_GET_VERSION,
            APDU_GET_CPLC,
            APDU_GET_CHALLENGE,
            "",
    };

    // OMAPI ressources
    private SEService mSeService;
    private Reader[] mReaders;
    private Reader mCurrentReader;
    private Session mCurrentSession;

    // UI resources
    private Button mRunButton;
    private Spinner mEseSpinner;
    private TextView mATRTextView;
    private Spinner mCommandSpinner;
    private EditText mApduEditText;
    private TextView mRespTextView;
    private TextView mTimeTextView;

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_main);

        // Init UI
        mApduEditText = findViewById(R.id.apdu);
        mATRTextView = findViewById(R.id.atr);
        mRespTextView = findViewById(R.id.resp);
        mTimeTextView = findViewById(R.id.time);
        mCommandSpinner = findViewById(R.id.command_spinner);
        mRunButton = findViewById(R.id.run);

        mCommandSpinner.setAdapter(new ArrayAdapter<>(this,
                android.R.layout.simple_spinner_item, COMMAND_ARRAY));
        mCommandSpinner.setOnItemSelectedListener(new AdapterView.OnItemSelectedListener() {
            public void onItemSelected(android.widget.AdapterView<?> parent, View view, int pos,
                                       long id) {
                String apdu = APDU_ARRAY[pos];
                mApduEditText.setText(apdu);
            }

            public void onNothingSelected(android.widget.AdapterView<?> parent) {
                mApduEditText.setText("");
            }
        });

        mEseSpinner = findViewById(R.id.eseSpinner);
        mEseSpinner.setOnItemSelectedListener(new AdapterView.OnItemSelectedListener() {
            @Override
            public void onItemSelected(AdapterView<?> parent, View view, int position, long id) {
                mATRTextView.setText("");
                mRunButton.setEnabled(false);
                if (!mSeService.isConnected()) {
                    // Wait service
                    return;
                }

                // Check if the SE is present
                if (!mReaders[position].isSecureElementPresent()) {
                    mATRTextView.setText(R.string.no_card_present);
                    return;
                }

                // Close every previous sessions with readers
                if (mCurrentReader != null) {
                    mCurrentReader.closeSessions();
                }

                mCurrentReader = mReaders[position];

                // Open a new session with the selected reader
                try {
                    Log.i(TAG, "Create Session from the selected reader...");
                    mCurrentSession = mCurrentReader.openSession();

                    // Retrieve the ATR of SE if available (Some SE doesn't have ATR)
                    if (mCurrentSession.getATR() != null) {
                        mATRTextView.setText(ByteArrayUtils.bytesToHexString(
                                mCurrentSession.getATR()));
                    } else {
                        mATRTextView.setText("N/A");
                    }
                    mRunButton.setEnabled(true);
                } catch (IOException e) {
                    Toast.makeText(MainActivity.this, "Fail to open reader",
                            Toast.LENGTH_LONG).show();
                    mCurrentReader = null;
                }
            }

            @Override
            public void onNothingSelected(AdapterView<?> parent) {
                if (mCurrentReader != null) {
                    mCurrentReader.closeSessions();
                }
                mATRTextView.setText("");
                mRunButton.setEnabled(false);
            }
        });

    }

    @Override
    protected void onResume() {
        super.onResume();
        // Reset UI
        mEseSpinner.setEnabled(false);
        mRunButton.setEnabled(false);

        // Restore connection with the SmartcardService
        try {
            Log.i(TAG, "creating SEService object");
            mSeService = new SEService(this, this);
        } catch (SecurityException e) {
            Log.e(TAG, "Binding not allowed, uses-permission " +
                    "org.simalliance.openmobileapi.SMARTCARD?");
        } catch (NullPointerException e) {
            Log.e(TAG, "Invalid context", e);
        }
    }

    @Override
    protected void onPause() {
        // Release resources
        if (mCurrentReader != null) {
            mCurrentReader.closeSessions();
            mCurrentSession = null;
        }
        if (mSeService != null && mSeService.isConnected()) {
            mSeService.shutdown();
        }
        mCurrentReader = null;

        super.onPause();
    }

    // Smartcard service Callback
    public void serviceConnected(SEService service) {
        Log.i(TAG, "Retrieve available readers...");
        mReaders = mSeService.getReaders();
        if (mReaders.length < 1) return;

        // Fill the spinner with the reader list
        List<String> spinnerArray =  new ArrayList<>();
        for (Reader reader : mReaders) {
            spinnerArray.add(reader.getName());
        }
        ArrayAdapter<String> adapter = new ArrayAdapter<>(
                this, android.R.layout.simple_spinner_item, spinnerArray);

        adapter.setDropDownViewResource(android.R.layout.simple_spinner_dropdown_item);
        mEseSpinner.setAdapter(adapter);
        mEseSpinner.setEnabled(true);
    }

    public void onRun(View v) {
        String hexString = mApduEditText.getEditableText().toString();
        byte[] apdu;
        byte[] resp;
        long start;
        long end;
        Channel channel = null;

        try {
            apdu = ByteArrayUtils.fromHexString(hexString);

            // Open a communication channel with the SE
            Log.i(TAG, "Create basic channel within the session...");
            // In case of select command, the AID has to be a parameter of the openBasicChannel
            if (apdu.length > 1 && apdu[1] == (byte)0xA4) {
                start = System.currentTimeMillis();
                channel = mCurrentSession.openBasicChannel(Arrays.copyOfRange(apdu, 5,
                        apdu.length));
                end = System.currentTimeMillis();
                if (channel == null) {
                    throw new IllegalStateException("Fail to open channel");
                }
                resp = channel.getSelectResponse();
            } else {
                channel = mCurrentSession.openBasicChannel(null);

                if (channel == null) {
                    // Maybe this SE supports logical channel, try with logical channel.
                    channel = mCurrentSession.openLogicalChannel(null);
                    if (channel == null) {
                        throw new IllegalStateException("Fail to open channel");
                    }
                }

                // send APDU
                Log.i(TAG, "APDU => " + ByteArrayUtils.bytesToHexString(apdu));
                start = System.currentTimeMillis();
                resp = channel.transmit(apdu);
                end = System.currentTimeMillis();
            }

            String s = ByteArrayUtils.bytesToHexString(resp, true);
            Log.i(TAG, "APDU <= " + s);
            mRespTextView.setText(s);
            mTimeTextView.setText(String.format(Locale.US, "%dms", (end - start)));
        } catch (IOException | IllegalStateException | IllegalArgumentException | SecurityException
                | NullPointerException | UnsupportedOperationException | NoSuchElementException e) {
            Log.e(TAG, "Processing error: ", e);
            Toast.makeText(this, e.toString(), Toast.LENGTH_LONG).show();
        } finally {
            if (channel != null) {
                channel.close();
            }
        }
    }

    public void onReset(View v) {
        mCommandSpinner.setSelection(0);
        mApduEditText.setText(APDU_ARRAY[0]);
        mATRTextView.setText(R.string.not_available);
        mRespTextView.setText(R.string.not_available);
        mTimeTextView.setText(R.string.not_available);
    }
}